// 2-Dim line integral convolution
// Andreas Hirczy <ahi@sbox.tu-graz.ac.at>
//

static char rcsid[]="$Id: lin_conv.cc,v 1.9 1998/01/21 15:38:10 ahi Exp $";

// Includes:
#include "mex.h"
#include "assert.h"
#include <unistd.h>
#include "math.h"


// Global variables: number of rows, number of columns
static int nrRows, nrCols;

// pointer to convolution matrices
double *vx, *vy;
double *oriNoise, *oriHue; // pointer to input  matrices
double *conNoise, *conHue; // pointer to output matrices
double *angle;

double h;         // stepsize for integrator
int    num;       // length of convolution kernel


// constants
const double Pi  = 3.14156;
const double Pi2 = 2.0 * Pi;

const double xx = 52.72077939; // angle for adjacent matrix elements 
const double xs = xx/sqrt(2);  // angle for diagonal matrix elements 
const double x2 = xx/2.0;


// Forwards
void lic_one(int row, int col);
int  lic_dir(int direction, int row, int col, double &val);


//////////////////////////////////
// access functions to matrices
//////////////////////////////////
inline double 
getElem(const double *aArr, const int aRow, const int aCol)
{
  assert((aRow >= 0)&&(aRow < nrRows));
  assert((aCol >= 0)&&(aCol < nrCols));

  return *(aArr + aCol*nrRows + aRow);
};

inline void
setElem(double *aArr, const int aRow, const int aCol, const double aVal)
{
  assert((aRow >= 0)&&(aRow < nrRows));
  assert((aCol >= 0)&&(aCol < nrCols));

  *(aArr + aCol*nrRows + aRow) = aVal;
};




///////////////////////////////////////////
// calculate LIC for every point in field
///////////////////////////////////////////
void 
lic_field(void)
{
  // Calculate angles
  double om;
  double v_x, v_y;
  angle = new double[nrRows*nrCols];
  for (int row=0; row<nrRows; row++)     // all rows
    for (int col=0; col<nrCols; col++)   // all columns
      {
	v_x = getElem(vx, row, col);
	v_y = getElem(vy, row, col);
	om = atan2(-v_y, v_x) * 180/Pi;	 // I had to use (-v_y) here?????
	setElem(angle, row, col, om);
      }

  // for every point 
  for (int row=0; row<nrRows; row++)
    for (int col=0; col<nrCols; col++)
	lic_one(row, col);


  
  free(angle);
}



/////////////////////////////////////
// calculate LIC for pixel at (i,j)
/////////////////////////////////////
void 
lic_one(int row, int col)
{
  int z;        // steps done
  double val;   // sum from linear convolution

  //  mexPrintf("LIC_ONE for %d/%d:", row+1, col+1);

  // considering the current origin
  z = 1;
  val = getElem(oriNoise, row, col);

  z += lic_dir(0, row, col, val);    // first going forward
  z += lic_dir(1, row, col, val);    // then back

  // normalize (boxfilter)
  val /= z; 

  //  mexPrintf(" --> value: %f from %d points.\n", val,z);

  setElem(conNoise, row, col, val);
}


int
lic_dir(int direction, int row, int col, double &val)
{
  assert((0 == direction) || (1 == direction));

  double omega; // direction obtained from vector field
  double om;    // direction we are to go to reach next point
  double err;   // error made by step

  int c;    // step counter
  for(c=0; c<num; c++)
    {
      // TODO: stop if velocity == 0
      
      // consider the angular error made by our last step
      omega =  getElem(angle, row, col) + direction*Pi + err;
      while(omega < 0)     // normalize angle;
	omega += 360;
      while(omega >= 360)
	omega -= 360;
      
      // calculate next point: looking for neighbour in direction omega
      // split at: 45, 135, 225, 315
      if ((omega <= x2) || (omega > 360.0-x2)){       // ---> right
	om = 0.0;
	col++;
      } else if (omega <= x2+xs){            // ---> up+right
	om = 45.0;
	col++;
	row--;
      } else if (omega <= x2+xs+xx){     // ---> up
	om = 90;
	row--;
      } else if (omega <= x2+2*xs+xx){  // ---> up+left
	om = 135;
	row--;
	col--;
      } else if (omega <= x2+2*xs+2*xx){    // ---> left
	om = 180;
	col--;
      } else if (omega <= x2+3*xs+2*xx){    // ---> down+left
	om = 225;
	row++;
	col--;
      } else if (omega <= x2+3*xs+3*xx){    // ---> down 
	om = 270;
	row++;
      } else if (omega <= x2+4*xs+3*xx){    // ---> down+right
	om = 315;
	row++;
	col++;
      }
      err = omega - om;
            
      // check boundaries
      if ((col<0) || (col>=nrCols) || (row<0) || (row>=nrRows))
	break;
      
      // accumulate sum
      val += getElem(oriNoise, row, col);
    }
  return c;
}




////////////////////////
// Interface to MATLAB
////////////////////////

#define IN_VX    prhs[0]
#define IN_VY    prhs[1]
#define IN_NOISE prhs[2]
#define IN_HUE   prhs[3]
#define IN_H     prhs[4]
#define IN_NUM   prhs[5]

#define OUT_LIC_NOISE plhs[0]
#define OUT_LIC_HUE   plhs[1]


void 
mexFunction(int nlhs,       mxArray *plhs[],
	    int nrhs, const mxArray *prhs[])
{

  // Check for proper number of arguments
  if (6 != nrhs)
    mexErrMsgTxt("Requires six input arguments.");
  if (2 != nlhs)
    mexErrMsgTxt("Requires two output arguments.");

  // Check dimensions (TODO: checking all matrices)
  nrRows = mxGetM(IN_NOISE);
  nrCols = mxGetN(IN_NOISE);

  mexPrintf("Matrix size: %d rows and %d columns\n", nrRows, nrCols);

 // Create matrices for the return arguments
  OUT_LIC_NOISE = mxCreateDoubleMatrix(nrRows, nrCols, mxREAL);
  OUT_LIC_HUE   = mxCreateDoubleMatrix(nrRows, nrCols, mxREAL);
  
  // Assign pointers to the various parameters
  vx       = mxGetPr(IN_VX);
  vy       = mxGetPr(IN_VY);
  oriNoise = mxGetPr(IN_NOISE);
  oriHue   = mxGetPr(IN_HUE);
  h        = mxGetScalar(IN_H);
  num      = int(floor(mxGetScalar(IN_NUM)));
  conNoise = mxGetPr(OUT_LIC_NOISE);
  conHue   = mxGetPr(OUT_LIC_HUE);

  // do the computations in a subroutine
  lic_field();

  return;
};

